"""Workflow Manager database code."""

from collections import namedtuple

from beeflow.common.db import bdb


class WorkflowInfo:
    """Workflow Info object."""

    def __init__(self, db_file):
        """Initialize Info and db file."""
        self.Info = namedtuple("Info", "id wfm_port tm_port sched_port num_workflows") # noqa Snake Case
        self.db_file = db_file

    def set_port(self, component, new_port):
        """Set port for the specified component."""
        stmt = f"UPDATE info set {component}_port=?"
        bdb.run(self.db_file, stmt, [new_port])

    def get_port(self, component):
        """Return port for the specified component."""
        # Need to add code here to make sure we chose a valid component.
        stmt = f"SELECT {component}_port FROM info"
        result = bdb.getone(self.db_file, stmt)
        port = result
        return port

    def increment_num_workflows(self):
        """Set workflow manager port."""
        stmt = "UPDATE info SET num_workflows = num_workflows + 1"
        bdb.run(self.db_file, stmt)

    def get_num_workflows(self):
        """Return scheduler port."""
        stmt = "SELECT num_workflows FROM info"
        result = bdb.getone(self.db_file, stmt)
        sched_port = result[0][0]
        return sched_port

    def get_info(self):
        """Return an info object containing port information."""
        stmt = "SELECT * FROM info"
        result = bdb.getone(self.db_file, stmt)
        info = self.Info(*result)
        return info


class Workflows:
    """Workflow database object."""

    def __init__(self, db_file):
        """Initialize Task, db_file, and Workflow object."""
        self.Task = namedtuple("Task", "id task_id workflow_id name resource state slurm_id") #noqa
        self.db_file = db_file
        self.Workflow = namedtuple("Workflow", "id workflow_id name state run_dir bolt_port http_port https_port gdb_pid") #noqa

    def get_workflow(self, workflow_id):
        """Return a workflow object."""
        stmt = "SELECT * FROM workflows WHERE workflow_id=?"
        result = bdb.getone(self.db_file, stmt, [workflow_id])
        workflow = self.Workflow(*result)
        return workflow

    def get_workflows(self):
        """Return a list of all the workflows."""
        stmt = "SELECT * FROM workflows"
        result = bdb.getall(self.db_file, stmt)
        workflows = [self.Workflow(*workflow) for workflow in result]
        return workflows

    def init_workflow(self, workflow_id, name, run_dir, bolt_port, http_port, https_port):
        """Insert a new workflow into the database."""
        stmt = """INSERT INTO workflows (workflow_id, name, state, run_dir,
                                         bolt_port, http_port, https_port, gdb_pid)
                  VALUES(?, ?, ?, ?, ?, ?, ?, ?);"""
        bdb.run(self.db_file, stmt, [workflow_id, name, 'Initializing', run_dir,
                                     bolt_port, http_port, https_port, -1])

    def delete_workflow(self, workflow_id):
        """Delete a workflow from the database."""
        stmt = "DELETE FROM workflows WHERE workflow_id=?"
        bdb.run(self.db_file, stmt, [workflow_id])

    def update_workflow_state(self, workflow_id, state):
        """Update the state in a workflow in the database."""
        stmt = "UPDATE workflows SET state=? WHERE workflow_id=?"
        bdb.run(self.db_file, stmt, [state, workflow_id])

    def get_workflow_state(self, workflow_id):
        """Return the bolt port associated with a workflow."""
        stmt = "SELECT state FROM workflows WHERE workflow_id=?"
        result = bdb.getone(self.db_file, stmt, [workflow_id])[0]
        state = result
        return state

    def add_task(self, task_id, workflow_id, name, state):
        """Add a task to the database associated with the workflow id specified."""
        stmt = "INSERT INTO tasks (task_id, workflow_id, name, resource, state,"\
               "slurm_id) VALUES(?, ?, ?, ?, ?, ?)"
        bdb.run(self.db_file, stmt, [task_id, workflow_id, name, "", state, -1])

    def delete_task(self, task_id, workflow_id):
        """Delete a task associed with a workflow."""
        stmt = "DELETE FROM tasks WHERE task_id=? AND workflow_id=?"
        bdb.run(self.db_file, stmt, [task_id, workflow_id])

    def update_task_state(self, task_id, workflow_id, state):
        """Update the state of a task."""
        stmt = "UPDATE tasks SET state=? WHERE task_id=? AND workflow_id=? "
        bdb.run(self.db_file, stmt, [state, task_id, workflow_id])

    def get_tasks(self, workflow_id):
        """Get all tasks associated with a particular workflow."""
        stmt = "SELECT * FROM tasks WHERE workflow_id=?"
        result = bdb.getall(self.db_file, stmt, [workflow_id])
        tasks = [self.Task(*task) for task in result]
        return tasks

    def get_task(self, task_id, workflow_id):
        """Get a task associed with a workflow."""
        stmt = "SELECT * FROM tasks WHERE task_id=? AND workflow_id=?"
        result = bdb.getone(self.db_file, stmt, [task_id, workflow_id])
        return result

    def get_bolt_port(self, workflow_id):
        """Return the bolt port associated with a workflow."""
        stmt = "SELECT bolt_port FROM workflows WHERE workflow_id=?"
        result = bdb.getone(self.db_file, stmt, [workflow_id])[0]
        bolt_port = result
        return bolt_port

    def get_gdb_pid(self, workflow_id):
        """Return the bolt port associated with a workflow."""
        stmt = "SELECT gdb_pid FROM workflows WHERE workflow_id=?"
        result = bdb.getone(self.db_file, stmt, [workflow_id])[0]
        gdb_pid = result
        return gdb_pid

    def update_gdb_pid(self, workflow_id, gdb_pid):
        """Update the gdb PID associated with a workflow."""
        stmt = "UPDATE workflows SET gdb_pid=? WHERE workflow_id=?"
        bdb.run(self.db_file, stmt, [gdb_pid, workflow_id])

    def get_run_dir(self, workflow_id):
        """Return the bolt port associated with a workflow."""
        stmt = "SELECT run_dir FROM workflows WHERE workflow_id=?"
        result = bdb.getone(self.db_file, stmt, [workflow_id])[0]
        run_dir = result
        return run_dir


class WorkflowDB:
    """Workflow manager database."""

    def __init__(self, db_file):
        """Initialize tables and db file."""
        self.db_file = db_file
        self._init_tables()

    def _init_tables(self):
        """Initialize the tables."""
        workflows_stmt = """CREATE TABLE IF NOT EXISTS workflows (
                                id INTEGER PRIMARY KEY,
                                -- Set workflow ID to unique.
                                workflow_id INTEGER UNIQUE,
                                name TEXT,
                                state TEST NOT NULL,
                                run_dir STR,
                                bolt_port INTEGER,
                                http_port INTEGER,
                                https_port INTEGER,
                                gdb_pid INTEGER);"""

        tasks_stmt = """CREATE TABLE IF NOT EXISTS tasks (
                        id INTEGER PRIMARY KEY,
                        task_id INTEGER UNIQUE,
                        workflow_id INTEGER NOT NULL,
                        name TEXT,
                        resource TEXT,
                        state TEXT,
                        slurm_id INTEGER,
                        FOREIGN KEY (workflow_id)
                            REFERENCES workflows (workflow_id)
                                ON DELETE CASCADE
                                ON UPDATE NO ACTION);"""

        info_stmt = """CREATE TABLE IF NOT EXISTS info (
                           id INTEGER PRIMARY KEY,
                           wfm_port INTEGER,
                           tm_port INTEGER,
                           sched_port INTEGER,
                           num_workflows INTEGER
                           );"""

        bdb.create_table(self.db_file, workflows_stmt)
        bdb.create_table(self.db_file, tasks_stmt)
        if not bdb.table_exists(self.db_file, 'info'):
            bdb.create_table(self.db_file, info_stmt)
            # insert a new workflow into the database
            stmt = """INSERT INTO info (wfm_port, tm_port, sched_port, num_workflows)
                                            VALUES(?, ?, ?, ?);"""
            bdb.run(self.db_file, stmt, [-1, -1, -1, 0])

    @property
    def workflows(self):
        """Get workflow info from the database."""
        return Workflows(self.db_file)

    @property
    def info(self):
        """Get workflow info from the database."""
        return WorkflowInfo(self.db_file)


def open_db(db_file):
    """Open a new database."""
    return WorkflowDB(db_file)



"""BEE configuration driver module."""

from configparser import ConfigParser
import getpass
import os
import platform
import random
import shutil
import sys
import textwrap
import typer

from beeflow.common.config_validator import ConfigValidator
from beeflow.common.cli import NaturalOrderGroup
from beeflow.common import validation
from beeflow.common.tab_completion import filepath_completion


# System specific path set up
HOME_DIR = os.path.expanduser('~/')
_SYSTEM = platform.system()
if _SYSTEM == 'Linux':
    CONF_DIR = os.path.join(HOME_DIR, '.config/beeflow')
    SYSCONFIG_FILE = '/etc/beeflow/bee.conf'
    USERCONFIG_FILE = os.path.expanduser('~/.config/beeflow/bee.conf')
elif _SYSTEM == 'Darwin':
    CONF_DIR = os.path.expanduser('~/Library/Application Support/beeflow')
    SYSCONFIG_FILE = '/Library/Application Support/beeflow/bee.conf'
    USERCONFIG_FILE = os.path.expanduser(
        '~/Library/Application Support/beeflow/bee.conf')
elif _SYSTEM == 'Windows':
    CONF_DIR = os.path.expandvars(r'%APPDATA%\beeflow')
    SYSCONFIG_FILE = None
    USERCONFIG_FILE = os.path.expandvars(r'%APPDATA%\beeflow\bee.conf')
else:
    raise RuntimeError(f'System "{_SYSTEM}" is not supported')

BEE_CONFIG = os.getenv('BEE_CONFIG')
if BEE_CONFIG is not None:
    USERCONFIG_FILE = BEE_CONFIG


class BeeConfig:
    r"""Class to manage and store all BEE configuration.

    All configuration values can be retrieved by using the get(section, key)
    class method. If those particular values are not in the config file, then
    defaults will be set, or a validation error raised, based on the validation code
    within the module.

    When new configuration keys are needed, they should be added to the
    validation code, along with information about the allowed input types, and
    general text for the user. If a key isn't added, but other code tries to
    access it, then an error will be raised letting the programmer know that the
    key needs to be added to the validation. This will hopefully help manage
    complexity and act as documentation as more keys are added.

    Configuration file locations by supported platform:

    Linux:
      sysconfig_file = '/etc/beeflow/bee.conf'

      userconfig_file = '~/.config/beeflow/bee.conf'

    MacOS:
      sysconfig_file = '/Library/Application Support/beeflow/bee.conf'

      userconfig_file = '~/Library/Application Support/beeflow/bee.conf'

    Windows:
      sysconfig_file = NOT SUPPORTED. Should be windows registry.

      userconfig_file = '%APPDATA%\\beeflow\\bee.conf'
    """

    CONFIG = None

    def __init__(self, **kwargs):
        """Do not use this constructor."""
        raise RuntimeError(
            'BeeConfig is a singleton class. Call BeeConfig.init() once to initialize.'
        )

    @classmethod
    def init(cls, userconfig=None, **_kwargs):
        """Initialize BeeConfig class.

        We check the platform and read in system and user configuration files.
        Note that this only needs to be called if one needs to initialize the
        config from a different file or with different keyword arguments. If
        so, then this must be called before any calls to bc.get() are made,
        since that call will initialize the config with default settings.
        """
        global USERCONFIG_FILE
        if cls.CONFIG is not None:
            return
        config = ConfigParser()
        if userconfig is not None:
            USERCONFIG_FILE = userconfig
        # Try and read the file
        try:
            with open(USERCONFIG_FILE, encoding='utf-8') as fp:
                config.read_file(fp)
        except FileNotFoundError:
            sys.exit('Configuration file does not exist! Please try running `beeflow config new`.')
        # remove default keys from the other sections
        default_keys = list(config['DEFAULT'])
        config = {sec_name: {key: config[sec_name][key] for key in config[sec_name]
                             if sec_name == 'DEFAULT' or key not in default_keys} # noqa
                  for sec_name in config}
        # Validate the config
        cls.CONFIG = VALIDATOR.validate(config)

    @classmethod
    def userconfig_path(cls):
        """Get the path of the user config."""
        return USERCONFIG_FILE

    @classmethod
    def get(cls, sec_name, opt_name):
        """Get a configuration value.

        If this throws, then the configuration value is missing from the
        definition. Initialize the config if not already initialized. Default
        values are built into the ConfigValidator class, so there is no need
        to specify a default or a fallback here.
        """
        if cls.CONFIG is None:
            cls.init()
        try:
            return cls.CONFIG[sec_name][opt_name] # noqa (this object is subscritable)
        except KeyError:
            raise RuntimeError(
                f'Option {sec_name}::{opt_name} was not found. Please contact '
                'a BEE developer or check the validation code in '
                'beeflow/common/config_driver.py.'
            ) from None

    @staticmethod
    def resolve_path(relative_path):
        """Resolve relative paths to absolute paths.

        :param relative_path: Input path. May include "../"
        :type relative_path: string, path to file
        """
        # Discard redundant paths, iff Windows, replace(\,/)
        # Assume "~" means home dir
        relative_path = os.path.expanduser(os.path.normpath(relative_path))
        # Resolve the true path (expand relative path refs)
        tmp = os.getcwd()
        if os.path.isdir(relative_path):
            os.chdir(relative_path)
            absolute_path = os.getcwd()
        else:
            # Get desired config file name
            filename = os.path.basename(relative_path)
            os.chdir(os.path.dirname(relative_path))
            absolute_path = '/'.join([os.getcwd(), filename])
        os.chdir(tmp)
        return absolute_path


def check_yes(msg):
    """Check for a y/n answer."""
    res = input(f'{msg} [y/n] ')
    return res.lower() == 'y'


def check_choice(msg, opts):
    """Ask the user to pick from opts."""
    while True:
        value = input(f'{msg} ({",".join(opts)})')
        if value in opts:
            return value
        print(f'ERROR: invalid option: {value}')


def join_path(*pargs):
    """Join multiple dirnames together to form a path."""
    path = pargs[0]
    for part in pargs[1:]:
        path = os.path.join(path, part)
    return path


def bee_workdir_init(path, _cur_opts):
    """BEE workdir init function.

    :param path: chosen path for the bee workdir
    :param _cur_opts: current chosen options form the config generator
    :return: initialized bee workdir path
    """
    path = os.path.expanduser(path)
    if os.path.exists(path):
        return path
    if not check_yes(f'Path "{path}" does not exist.\nWould you like to create it?'):
        return path
    os.makedirs(path)
    return path


def filepath_completion_input(*pargs, **kwargs):
    """Input files/paths with tab completion."""
    with filepath_completion():
        return input(*pargs, **kwargs)


# Below is the definition of all bee config options, defaults and requirements.
# This will be used to validate config files on loading them in the BeeConfig
# singleton class above.
#
# Set up default ports, get parent's pid to offset ports.
# Windows note: uid method better but not available in Windows
if platform.system() == 'Windows':
    # Would prefer something like a uid for windows.
    OFFSET = os.getppid() % 100
else:
    OFFSET = os.getuid() % 100
DEFAULT_BOLT_PORT = 7687 + OFFSET
DEFAULT_HTTP_PORT = 7474 + OFFSET
DEFAULT_HTTPS_PORT = 7473 + OFFSET

DEFAULT_WFM_PORT = 5000 + OFFSET
DEFAULT_TM_PORT = 5050 + OFFSET
DEFAULT_SCHED_PORT = 5100 + OFFSET

DEFAULT_BEE_WORKDIR = join_path(HOME_DIR, '.beeflow')
USER = getpass.getuser()
# Create the validator
VALIDATOR = ConfigValidator('BEE configuration file and validation information.')
VALIDATOR.section('DEFAULT', info='Default bee.conf configuration section.')
VALIDATOR.option('DEFAULT', 'bee_workdir', info='main BEE workdir',
                 attrs={'default': DEFAULT_BEE_WORKDIR}, validator=validation.make_dir)
VALIDATOR.option('DEFAULT', 'workload_scheduler', choices=('Slurm', 'LSF', 'Flux', 'Simple'),
                 info='backend workload scheduler to interact with ')
VALIDATOR.option('DEFAULT', 'use_archive', validator=validation.bool_, attrs={'default': True},
                 info='use the BEE archiving functinality')
VALIDATOR.option('DEFAULT', 'neo4j_image', validator=validation.file_,
                 info='neo4j container image',
                 attrs={'input': filepath_completion_input})
VALIDATOR.option('DEFAULT', 'redis_image', validator=validation.file_,
                 info='redis container image',
                 attrs={'input': filepath_completion_input})
VALIDATOR.option('DEFAULT', 'max_restarts', validator=int,
                 attrs={'default': 3},
                 info='max number of times beeflow will restart a component on failure')
# Workflow Manager
VALIDATOR.section('workflow_manager', info='Workflow manager section.')
# Task manager
VALIDATOR.section('task_manager',
                  info='Task manager configuration and config of container to use.')
VALIDATOR.option('task_manager', 'container_runtime', attrs={'default': 'Charliecloud'},
                 choices=('Charliecloud', 'Singularity'),
                 info='container runtime to use for configuration')
VALIDATOR.option('task_manager', 'runner_opts', attrs={'default': ''},
                 info='special runner options to pass to the runner opts')
VALIDATOR.option('task_manager', 'background_interval', attrs={'default': 5},
                 validator=int,
                 info='interval at which the task manager processes queues and updates states')

# Note: The special attrs keyword can include anything. One use case is for
# storing a special 'init' function that can be used to initialize the
# parameter when a user is generating a new configuration.

# Charliecloud (depends on task_manager::container_runtime == Charliecloud)
VALIDATOR.section('charliecloud', info='Charliecloud configuration section.',
                  depends_on=('task_manager', 'container_runtime', 'Charliecloud'))
VALIDATOR.option('charliecloud', 'image_mntdir', attrs={'default': join_path('/tmp', USER)},
                 info='Charliecloud mount directory', validator=validation.make_dir)
# General job requirements
VALIDATOR.section('job', info='General job requirements')
VALIDATOR.option('job', 'default_account', validator=lambda val: val.strip(),
                 info='default account to launch jobs with (leave blank if none)')
VALIDATOR.option('job', 'default_time_limit', validator=validation.time_limit,
                 info='default account time limit (leave blank if none)')
VALIDATOR.option('job', 'default_partition', validator=lambda val: val.strip(),
                 info='default partition to run jobs on (leave blank if none)')


def validate_chrun_opts(opts):
    """Ensure that chrun_opts don't contain options that'll conflict with BEE."""
    args = opts.split()
    if '--cd' in args or '-c' in args:
        raise ValueError('Initial working directory option (-c or --cd) conflicts with BEE. '
                         'Please make sure that you choose the correct working directory on '
                         'workflow submission.')
    return opts


VALIDATOR.option('charliecloud', 'chrun_opts', attrs={'default': '--home'},
                 validator=validate_chrun_opts,
                 info='extra options to pass to ch-run')
VALIDATOR.option('charliecloud', 'setup', attrs={'default': ''},
                 info='extra Charliecloud setup to put in a job script')
# Graph Database
VALIDATOR.section('graphdb', info='Main graph database configuration section.')
VALIDATOR.option('graphdb', 'hostname', attrs={'default': 'localhost'},
                 info='hostname of database')
VALIDATOR.option('graphdb', 'dbpass', attrs={'default': 'password'}, info='password for database')
VALIDATOR.option('graphdb', 'bolt_port', attrs={'default': DEFAULT_BOLT_PORT}, validator=int,
                 info='port used for the BOLT API')
VALIDATOR.option('graphdb', 'http_port', attrs={'default': DEFAULT_HTTP_PORT}, validator=int,
                 info='HTTP port used for the graph database')
VALIDATOR.option('graphdb', 'https_port', attrs={'default': DEFAULT_HTTPS_PORT},
                 info='HTTPS port used for the graph database')
VALIDATOR.option('graphdb', 'gdb_image_mntdir', attrs={'default': join_path('/tmp', USER)},
                 info='graph database image mount directory', validator=validation.make_dir)
VALIDATOR.option('graphdb', 'sleep_time', validator=int, attrs={'default': 10},
                 info='how long to wait for the graph database to come up (this can take a while, '
                      'depending on the system)')
# Builder
VALIDATOR.section('builder', info='General builder configuration section.')
VALIDATOR.option('builder', 'deployed_image_root', attrs={'default': '/tmp'},
                 info='where to deploy container images', validator=validation.make_dir)
VALIDATOR.option('builder', 'container_output_path', attrs={'default': '/tmp'},
                 info='container output path', validator=validation.make_dir)
VALIDATOR.option('builder', 'container_archive',
                 attrs={'default': join_path(DEFAULT_BEE_WORKDIR, 'container_archive')},
                 info='container archive location')
VALIDATOR.option('builder', 'container_type', attrs={'default': 'charliecloud'},
                 info='container type to use')
# Slurmrestd (depends on DEFAULT:workload_scheduler == Slurm)
VALIDATOR.section('slurm', info='Configuration section for Slurm.',
                  depends_on=('DEFAULT', 'workload_scheduler', 'Slurm'))
VALIDATOR.option('slurm', 'use_commands', validator=validation.bool_,
                 attrs={'default': shutil.which('slurmrestd') is None},
                 info='if set, use slurm cli commands instead of slurmrestd')
DEFAULT_SLURMRESTD_SOCK = join_path('/tmp', f'slurm_{USER}_{random.randint(1, 10000)}.sock')
VALIDATOR.option('slurm', 'openapi_version', attrs={'default': 'v0.0.37'},
                 info='openapi version to use for slurmrestd')
# Scheduler
VALIDATOR.section('scheduler', info='Scheduler configuration section.')
SCHEDULER_ALGORITHMS = ('fcfs', 'backfill', 'sjf')
VALIDATOR.option('scheduler', 'algorithm', attrs={'default': 'fcfs'}, choices=SCHEDULER_ALGORITHMS,
                 info='scheduling algorithm to use')
VALIDATOR.option('scheduler', 'default_algorithm', attrs={'default': 'fcfs'},
                 choices=SCHEDULER_ALGORITHMS,
                 info=('default algorithm to use'))


def print_wrap(text, next_line_indent=''):
    """Print while wrapping lines to make the output easier to read."""
    for line in textwrap.wrap(text, width=80, subsequent_indent=next_line_indent):
        print(line)


class ConfigGenerator:
    """Config generator class."""

    def __init__(self, fname, validator):
        """Construct the config generator."""
        self.fname = fname
        self.validator = validator
        self.sections = {}

    def choose_values(self):
        """Choose configuration values based on user input."""
        dirname = os.path.dirname(self.fname)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        print(f'Creating a new config: "{self.fname}".')
        print()
        print_wrap('This will walk you through creating a new configuration for BEE. '
                   'Note that you will only be required to enter values for options '
                   'without defaults. Please take a look at the other options and '
                   'their defaults before running BEE.')
        print()
        print('Please enter values for the following sections and options:')
        # Let the user choose values for each required attribute
        for sec_name, section in self.validator.sections:
            # Determine if this section is valid under the current configuration
            if not self.validator.is_section_valid(self.sections, sec_name):
                continue
            self.sections[sec_name] = {}
            printed = False
            for opt_name, option in self.validator.options(sec_name):
                # Print the section name if it hasn't already been printed.
                if not printed:
                    print()
                    print(f'## {sec_name}')
                    print()
                    print_wrap(section.info)
                    print()
                    printed = True

                # Check for a default value
                if option.attrs is not None and 'default' in option.attrs:
                    default = option.attrs['default']
                    if option.attrs is not None and 'init' in option.attrs:
                        value = option.attrs['init'](default, self.sections)
                    else:
                        value = option.validate(default)
                    print(f'Setting option "{opt_name}" to default value "{value}".')
                    print()
                    self.sections[sec_name][opt_name] = value
                    continue

                # Input and validate the value
                value = self._input_loop(opt_name, option)

                print()
                self.sections[sec_name][opt_name] = value
        return self

    def _input_loop(self, opt_name, option):
        """Run the input-validation loop."""
        # Check if there's a special input function (this allows for tab completion)
        inp = (option.attrs['input'] if option.attrs is not None
               and 'input' in option.attrs else input)
        # Start of input loop
        value = None
        while value is None:
            print_wrap(f'{opt_name} - {option.info}')
            if option.choices is not None:
                print(f'(allowed values: {",".join(option.choices)})')
            value = inp(f'{opt_name}: ')
            # Call the init function if there is one
            if option.attrs is not None and 'init' in option.attrs:
                option.attrs['init'](value, self.sections)
            # Validate the input
            try:
                option.validate(value)
            except ValueError as err:
                print('ERROR:', err)
                value = None
        return value

    def save(self):
        """Save the config to a file."""
        print()
        print('The following configuration options were chosen:')
        for sec_name, section in self.sections.items():
            print()
            print(f'[{sec_name}]')
            for opt_name in section:
                print(f'{opt_name} = {section[opt_name]}')
        print()
        ans = input('Would you like to save this config? [y/n] ')
        if ans.lower() != 'y':
            print('Quitting without saving')
            return
        try:
            with open(self.fname, 'w', encoding='utf-8') as fp:
                print('# BEE Configuration File', file=fp)
                for sec_name, section in self.sections.items():
                    if not section:
                        continue
                    print(file=fp)
                    print(f'[{sec_name}]'.format(sec_name), file=fp)
                    for opt_name in section:
                        print(f'{opt_name} = {section[opt_name]}', file=fp)
        except FileNotFoundError:
            print('Configuration file does not exist!')
        print(70 * '#')
        print('Before running BEE, check defaults in the configuration file:',
              f'\n\t{self.fname}',
              '\n ** See documentation for values you should refrain from editing! **',
              '\n ** Include job options (such as account) required for this system.**')
        print('\n(Try `beeflow config info` to see more about each option)')
        print(70 * '#')


app = typer.Typer(no_args_is_help=True, add_completion=False, cls=NaturalOrderGroup)


@app.command()
def validate(path: str = typer.Argument(default=USERCONFIG_FILE,
                                        help='Path to config file')):
    """Validate an existing configuration file."""
    BeeConfig.init(path)


@app.command()
def info():
    """Display some info about bee.conf's various options."""
    print('# BEE Configuration')
    print()
    print_wrap(VALIDATOR.description)
    print()
    for sec_name, section in VALIDATOR.sections:
        print(f'## {sec_name}')
        if section.depends_on is not None:
            print()
            deps = section.depends_on
            print_wrap(f'*only required if {deps[0]}::{deps[1]} == "{deps[2]}"*')
        print()
        print_wrap(f'{section.info}')
        print()
        for opt_name, option in VALIDATOR.options(sec_name):
            print_wrap(f'* {opt_name} - {option.info}', '  ')
            if option.choices is not None:
                print(f'\t* allowed values: {",".join(option.choices)}')
        print()


@app.command()
def new(path: str = typer.Argument(default=USERCONFIG_FILE,
                                   help='Path to new config file')):
    """Create a new config file."""
    if os.path.exists(path):
        if check_yes(f'Path "{path}" already exists.\nWould you like to save a copy of it?'):
            i = 1
            backup_path = f'{path}.{i}'
            while os.path.exists(backup_path):
                i += 1
                backup_path = f'{path}.{i}'
            shutil.copy(path, backup_path)
            print(f'Saved old config to "{backup_path}".')
            print()
    ConfigGenerator(path, VALIDATOR).choose_values().save()


@app.command()
def show(path: str = typer.Argument(default=USERCONFIG_FILE,
                                    help='Path to config file')):
    """Show the contents of bee.conf."""
    if not os.path.exists(path):
        print('The bee.conf does not exist yet. Please run `beeflow config new`.')
        return
    print(f'# {path}')
    with open(path, encoding='utf-8') as fp:
        print(fp.read(), end='')
# Ignore C901: "'ConfigGenerator.choose_values' is too complex" - I disagree, if
#              it's just based on LOC, then there are a number `print()` functions
#              that are increasing the line count
# pylama:ignore=C901



"""Task Manager app and api set up code.

Submits, cancels and monitors states of tasks.
Communicates status to the Work Flow Manager, through RESTful API.
"""
import atexit
import sys
from apscheduler.schedulers.background import BackgroundScheduler
from flask import Flask, jsonify, make_response
from beeflow.common.api import BeeApi
from beeflow.task_manager.task_submit import TaskSubmit
from beeflow.task_manager.task_actions import TaskActions
from beeflow.task_manager.background import process_queues
from beeflow.common.config_driver import BeeConfig as bc


def create_app():
    """Create the flask app and add the REST endpoints for the TM."""
    app = Flask(__name__)
    api = BeeApi(app)

    # Endpoints
    api.add_resource(TaskSubmit, '/bee_tm/v1/task/submit/')
    api.add_resource(TaskActions, '/bee_tm/v1/task/')

    @app.route('/status')
    def get_status():
        """Report the current status of the Task Manager."""
        return make_response(jsonify(stauts='up'), 200)

    # Start the background scheduler and make sure it gets cleaned up
    if "pytest" not in sys.modules:
        scheduler = BackgroundScheduler({'apscheduler.timezone': 'UTC'})
        scheduler.add_job(func=process_queues, trigger="interval",
                          seconds=bc.get('task_manager', 'background_interval'))
        scheduler.start()

        # This kills the scheduler when the process terminates
        # so we don't accidentally leave a zombie process
        atexit.register(scheduler.shutdown)

    return app



"""BEE integration tests."""
from pathlib import Path
import os
import shutil
import sys
import uuid
import typer

from beeflow.common.integration import utils
from beeflow.common.integration import generated_workflows
from beeflow.common.config_driver import BeeConfig as bc


TEST_RUNNER = utils.TestRunner()


@TEST_RUNNER.add()
def copy_container(outer_workdir):
    """Prepare, check results of using `beeflow:copyContainer` then do cleanup."""
    # `beeflow:copyContainer` workflow
    container_path = f'/tmp/copy_container-{uuid.uuid4().hex}.tar.gz'
    container_name = 'copy-container'
    workdir = os.path.join(outer_workdir, uuid.uuid4().hex)
    os.makedirs(workdir)
    container = utils.Container(container_name, generated_workflows.DOCKER_FILE_PATH,
                                container_path)
    workflow_path = os.path.join(outer_workdir, f'bee-cc-workflow-{uuid.uuid4().hex}')
    main_input = 'copy_container'
    docker_requirement = {
        'beeflow:copyContainer': container_path,
    }
    main_cwl, job_file = generated_workflows.builder_workflow(workflow_path, docker_requirement,
                                                              main_input)
    workflow = utils.Workflow('copy-container', workflow_path, main_cwl=main_cwl,
                              job_file=job_file, workdir=workdir,
                              containers=[container])
    yield [workflow]
    utils.check_completed(workflow)
    # Ensure the output file was created
    path = os.path.join(workdir, main_input)
    utils.check_path_exists(path)
    os.remove(path)
    # Ensure that the container has been copied into the archive
    container_archive = bc.get('builder', 'container_archive')
    basename = os.path.basename(container_path)
    path = os.path.join(container_archive, basename)
    utils.check_path_exists(path)
    os.remove(path)
    utils.ch_image_delete(container_name)


@TEST_RUNNER.add()
def use_container(outer_workdir):
    """Prepare, check results of using `beeflow:useContainer` and clean up."""
    # `beeflow:useContainer` workflow
    container_path = os.path.join(outer_workdir, f'use_container-{uuid.uuid4().hex}.tar.gz')
    container_name = 'use-container'
    container_workdir = os.path.join(outer_workdir, uuid.uuid4().hex)
    os.makedirs(container_workdir)
    container = utils.Container(container_name, generated_workflows.DOCKER_FILE_PATH,
                                container_path)
    workflow_path = os.path.join(outer_workdir, f'bee-use-ctr-workflow-{uuid.uuid4().hex}')
    main_input = 'use_ctr'
    docker_requirement = {
        'beeflow:useContainer': container_path,
    }
    main_cwl, job_file = generated_workflows.builder_workflow(
        workflow_path,
        docker_requirement,
        main_input,
    )
    workflow = utils.Workflow('use-container', workflow_path, main_cwl=main_cwl,
                              job_file=job_file, workdir=container_workdir,
                              containers=[container])
    yield [workflow]
    utils.check_completed(workflow)
    path = os.path.join(container_workdir, main_input)
    utils.check_path_exists(path)
    # This container should not have been copied into the container archive
    container_archive = bc.get('builder', 'container_archive')
    basename = os.path.basename(container_path)
    path = Path(container_archive, basename)
    utils.ci_assert(
        not path.exists(),
        f'the container "{basename}" was copied into the container archive'
    )
    utils.ch_image_delete(container_name)


@TEST_RUNNER.add()
def docker_file(outer_workdir):
    """Ensure that the `dockerFile` example ran properly and then clean up."""
    # `dockerFile` workflow
    workflow_path = os.path.join(outer_workdir, f'bee-df-workflow-{uuid.uuid4().hex}')
    # Copy the Dockerfile to the workdir path
    os.makedirs(workflow_path)
    workdir = os.path.join(outer_workdir, uuid.uuid4().hex)
    os.makedirs(workdir)
    shutil.copy(generated_workflows.DOCKER_FILE_PATH, os.path.join(workflow_path, 'Dockerfile'))
    container_name = 'docker_file_test'
    docker_requirement = {
        'dockerFile': 'Dockerfile',
        'beeflow:containerName': container_name,
    }
    main_input = 'docker_file'
    main_cwl, job_file = generated_workflows.builder_workflow(
        workflow_path,
        docker_requirement,
        main_input,
    )
    workflow = utils.Workflow('docker-file', workflow_path, main_cwl=main_cwl,
                              job_file=job_file, workdir=workdir, containers=[])
    yield [workflow]
    utils.check_completed(workflow)
    path = os.path.join(workdir, main_input)
    utils.check_path_exists(path)
    # The container should have been copied into the archive
    container_archive = bc.get('builder', 'container_archive')
    tarball = f'{container_name}.tar.gz'
    path = os.path.join(container_archive, tarball)
    utils.check_path_exists(path)
    os.remove(path)
    # Check that the container is listed
    images = utils.ch_image_list()
    utils.ci_assert(container_name in images,
                    f'cannot find expected container "{container_name}"')
    utils.ch_image_delete(container_name)


@TEST_RUNNER.add()
def docker_pull(outer_workdir):
    """Prepare, then check that the `dockerPull` option was successful."""
    # `dockerPull` workflow
    workflow_path = os.path.join(outer_workdir, f'bee-dp-workflow-{uuid.uuid4().hex}')
    workdir = os.path.join(outer_workdir, uuid.uuid4().hex)
    os.makedirs(workdir)
    container_name = generated_workflows.BASE_CONTAINER
    docker_requirement = {
        'dockerPull': container_name,
    }
    main_input = 'docker_pull'
    main_cwl, job_file = generated_workflows.builder_workflow(
        workflow_path,
        docker_requirement,
        main_input,
    )
    workflow = utils.Workflow('docker-pull', workflow_path, main_cwl=main_cwl,
                              job_file=job_file, workdir=workdir, containers=[])
    yield [workflow]
    utils.check_completed(workflow)
    path = os.path.join(workdir, main_input)
    utils.check_path_exists(path)
    # Check that the image tarball is in the archive
    container_archive = bc.get('builder', 'container_archive')
    path = os.path.join(container_archive, f'{container_name}.tar.gz')
    utils.check_path_exists(path)
    os.remove(path)
    # Commenting the below out for now; looks like Charliecloud 0.32 isn't
    # showing base containers for some reason?
    # Check for the image with `ch-image list`
    # images = utils.ch_image_list()
    # utils.ci_assert(container_name in images,
    #                 f'could not find expected container "{container_name}"')
    # utils.ch_image_delete(container_name)


@TEST_RUNNER.add()
def multiple_workflows(outer_workdir):
    """Test running three different workflows at the same time."""
    output_file = 'output_file'
    workflow_data = []
    workflows = []
    for i in range(3):
        workdir = os.path.join(outer_workdir, uuid.uuid4().hex)
        os.makedirs(workdir)
        workflow_path = Path(outer_workdir, uuid.uuid4().hex)
        main_cwl, job_file = generated_workflows.simple_workflow(workflow_path, output_file)
        workflow = utils.Workflow(f'multi-workflow-{i}', workflow_path,
                                  main_cwl=main_cwl, job_file=job_file,
                                  workdir=workdir, containers=[])
        workflow_data.append({
            'workdir': workdir,
            'workflow_path': workflow_path,
        })
        workflows.append(workflow)
    yield workflows
    for workflow in workflows:
        utils.check_completed(workflow)
    for wfl_info in workflow_data:
        workdir = wfl_info['workdir']
        workflow_path = wfl_info['workflow_path']
        # Each test should have touched this file
        path = Path(workdir, output_file)
        utils.check_path_exists(path)


@TEST_RUNNER.add()
def build_failure(outer_workdir):
    """Test running a workflow with a bad container."""
    workdir = os.path.join(outer_workdir, uuid.uuid4().hex)
    os.makedirs(workdir)
    workflow = utils.Workflow('build-failure', 'ci/test_workflows/build-failure',
                              main_cwl='workflow.cwl', job_file='input.yml',
                              workdir=workdir, containers=[])
    yield [workflow]
    utils.check_workflow_failed(workflow)
    # Only one task
    task_state = workflow.task_states[0][2]
    utils.ci_assert(task_state == 'BUILD_FAIL',
                    f'task was not in state BUILD_FAIL as expected: {task_state}')


@TEST_RUNNER.add(ignore=True)
def checkpoint_restart(outer_workdir):
    """Test the clamr-ffmpeg checkpoint restart workflow."""
    workdir = os.path.join(outer_workdir, uuid.uuid4().hex)
    os.makedirs(workdir)
    workflow = utils.Workflow('checkpoint-restart',
                              'ci/test_workflows/clamr-wf-checkpoint',
                              main_cwl='clamr_wf.cwl', job_file='clamr_job.yml',
                              workdir=workdir, containers=[])
    yield [workflow]
    utils.check_completed(workflow)
    # Check for the movie file
    path = Path(workdir, 'CLAMR_movie.mp4')
    utils.check_path_exists(path)


@TEST_RUNNER.add(ignore=True)
def checkpoint_restart_failure(outer_workdir):
    """Test a checkpoint restart workflow that continues past 'num_retries'."""
    workdir = os.path.join(outer_workdir, uuid.uuid4().hex)
    os.makedirs(workdir)
    workflow = utils.Workflow('checkpoint-too-long',
                              'ci/test_workflows/checkpoint-too-long',
                              main_cwl='workflow.cwl', job_file='input.yml',
                              workdir=workdir, containers=[])
    yield [workflow]
    utils.check_workflow_failed(workflow)


def test_input_callback(arg):
    """Parse a list of tests separated by commas."""
    return arg.split(',') if arg is not None else None


def main(tests = typer.Option(None, '--tests', '-t',  # noqa (conflict on '=' sign)
                              callback=test_input_callback,
                              help='tests run as comma-separated string'),
         show_tests: bool = typer.Option(False, '--show-tests', '-s',  # noqa (conflict on '=' sign)
                                         help='show a list of all tests'),
         timeout: int = typer.Option(utils.TIMEOUT, '--timeout',
                                     help='workflow timeout in seconds')):
    """Launch the integration tests."""
    if show_tests:
        print('INTEGRATION TEST CASES:')
        for test_name, ignore in TEST_RUNNER.test_details():
            print('*', test_name, '(ignored)' if ignore else '')
        return
    generated_workflows.init()
    TEST_RUNNER.timeout = timeout
    # Run the workflows and then show completion results
    ret = TEST_RUNNER.run(tests)
    # ret = test_workflows(WORKFLOWS)
    # General clean up
    os.remove(generated_workflows.DOCKER_FILE_PATH)
    sys.exit(ret)
# Ignore W0231: This is a user-defined exception and I don't think we need to call
#               __init__ on the base class.
# pylama:ignore=W0231



"""Workflow Generator."""
from scriptcwl import WorkflowGenerator

with WorkflowGenerator() as wf:
    wf.load(steps_dir='/Users/raginigupta/PythonCodes/cwl')

    num1 = wf.add_input(num1='int')
    num2 = wf.add_input(num2='int')

    answer1 = wf.add(x=num1, y=num2)
    answer2 = wf.multiply(x=answer1, y=num2)

    wf.add_outputs(final_answer=answer2)

    wf.save('add_multiply_example_workflow.cwl')
